import os
import pickle
import hashlib
from pathlib import Path
from typing import Union

DEFAULT_CACHE_DIR = os.path.join(Path.home(), '.cache')
BUF_SIZE = 65536  # lets read stuff in 64kb chunks!


class PersistentCache:
    class Key:
        key: any
        keytype: str

        __hexdigest_value: str

        def __init__(self, key, keytype: str):
            assert isinstance(keytype, str) and keytype in ['str', 'file']
            self.key = key
            self.keytype = keytype
            self.__hexdigest_value = self._hexdigest()

        @property
        def hexdigest(self) -> str:
            return self.__hexdigest_value

        def _hexdigest(self) -> str:
            raise NotImplementedError

        def __hash__(self):
            return hash(hash(self.key) ^ hash(self.keytype) ^ hash(self.hexdigest))

        def __eq__(self, other):
            return hash(self) == hash(other)

    class StrKey(Key):
        def __init__(self, key):
            assert isinstance(key, str)
            super().__init__(key, 'str')

        def _hexdigest(self) -> str:
            return hex(hash(self.key))

    class FileKey(Key):
        def __init__(self, key):
            assert isinstance(key, str)
            if not os.path.isfile(key):
                raise FileNotFoundError(key)

            super().__init__(os.path.realpath(key), 'file')

        def _hexdigest(self) -> str:
            sha1 = hashlib.sha1()

            with open(self.key, 'rb') as f:
                while True:
                    data = f.read(BUF_SIZE)
                    if not data:
                        break
                    sha1.update(data)
            return sha1.hexdigest()

    class _Entry:
        key: any
        value: any

        def __init__(self, key, value):
            self.key = key
            self.value = value

    cache_dir: str
    keys: set

    def __locate_value(self, key: Key) -> str:
        return os.path.join(self.cache_dir, '__cache_entry_' + key.hexdigest)

    def __init__(self, cache_dir=DEFAULT_CACHE_DIR):
        self.cache_dir = cache_dir
        self.keys = set()

        if not os.path.isdir(cache_dir):
            raise NotADirectoryError(cache_dir)

        for file in os.listdir(self.cache_dir):
            path = os.path.join(self.cache_dir, file)
            if not os.path.isfile(path):
                continue

            try:
                with open(path, 'rb') as f:
                    e: PersistentCache._Entry = pickle.load(f)
                assert isinstance(e, PersistentCache._Entry)
            except Exception as e:
                print(e)
                continue

            self.keys.add(e.key)

    def lookup(self, key: Key) -> any:
        if key not in self:
            raise KeyError(key)

        value_path = self.__locate_value(key)
        with open(value_path, 'rb') as f:
            e: PersistentCache._Entry = pickle.load(f)
        return e.value

    def update(self, key: Key, value):
        value_path = self.__locate_value(key)
        with open(value_path, 'wb') as f:
            pickle.dump(PersistentCache._Entry(key, value), f)

        matches = [*filter(lambda e: e == key, self.keys)]
        if matches:
            assert len(matches) == 1
            self.remove(matches[0])
        self.keys.add(key)

    def remove(self, key: Key):
        value_path = self.__locate_value(key)
        if os.path.isfile(value_path):
            os.remove(value_path)
        self.keys.remove(key)

    def __contains__(self, item):
        return item in self.keys



import getpass
import os
import shutil
import tempfile
from os import path as _p

import sh

from .pyshutil import compile_shargs


def __build(build_context, fg=False):
    tmpdir = tempfile.TemporaryDirectory()

    iidfile = _p.join(tmpdir.name, '__iid')
    try:
        sh.docker.build(build_context, iidfile=iidfile, _fg=fg)
    except Exception:
        raise Exception(f'failed to build a docker image with build context at {build_context}')

    with open(iidfile, 'r') as f:
        return f.read()


def __create(image_id: str, commands: str, volumes=None):
    tmpdir = tempfile.TemporaryDirectory()

    cidfile = _p.join(tmpdir.name, '__cid')

    if not volumes:
        v_opts = []
    elif isinstance(volumes, str):
        v_opts = [f'-v={volumes}']
    elif isinstance(volumes, list):
        v_opts = [*map(lambda o: f'-v={o}', volumes)]
    else:
        raise Exception

    sh.docker.create(f'--cidfile={cidfile}',
                     *v_opts,
                     '-i',
                     '--rm',  # Automatically remove the container when it exits
                     image_id,
                     '/bin/bash',
                     c=commands,
                     )
    with open(cidfile, 'r') as f:
        return f.read()


def dsh(*args,
        _volumes=None,
        _root=False,
        _auto_remove=True,
        _verbose=False,
        _cwd=None,
        _build_context=None,
        **kwargs,
        ):

    if (not _build_context) or (not _p.isdir(_build_context)):
        raise Exception('proper _build_context directory must be supplied')

    if not _cwd:
        _cwd = _p.curdir

    _cwd = _p.realpath(_cwd)

    commands, shkwargs = compile_shargs(*args, **kwargs)

    commands = ' '.join(commands)

    image_id = __build(_build_context, fg=_verbose)
    if not image_id:
        raise Exception('failed to build image')

    if _root:
        home = '/root'
        commands = f'cd {home} && {commands}'
    else:
        username = getpass.getuser()
        uid = os.getuid()
        gid = os.getgid()
        home = f'/home/{username}'

        commands = (
            f'if [ "$(id -u {username} > /dev/null 2>&1; echo $?)" == 0 ]; then userdel {username}; fi && '
            f'groupadd -g {gid} {username}; '
            f'useradd -l -u {uid} -g $(getent group {gid} | cut -d: -f1) {username} && '
            f'install -d -m 0755 -o {username} -g $(getent group {gid} | cut -d: -f1) {home} && '
            f'chown -R {uid}:{gid} {home} && '
            f'cd {home} && su {username} -c "/bin/bash -c \\"{commands}\\""'
        )

    # print(commands)

    container_id = __create(image_id, commands, f'{_cwd}:{home}:rw')
    if not container_id:
        raise Exception('failed to create container')

    fg = shkwargs['_fg'] if '_fg' in shkwargs else False

    if fg:
        return sh.docker.start(container_id, i=True, **shkwargs)
    else:
        return sh.docker.start(container_id, a=True, **shkwargs)



from os import path as _p
from unittest import TestCase

from yautil import dsh
from tempfile import TemporaryDirectory


class TestDocker(TestCase):
    tmpdir: TemporaryDirectory

    def setUp(self):
        self.tmpdir = TemporaryDirectory()

    def tearDown(self) -> None:
        self.tmpdir.cleanup()

    def test_hello(self):
        dockerfile = _p.join(self.tmpdir.name, 'Dockerfile')
        with open(dockerfile, 'w+') as f:
            f.write('FROM ubuntu')
        sout = str(dsh('echo', '-n', 'hello', _build_context=self.tmpdir.name))
        assert sout == 'hello'



import setuptools


deps = [
    'sh',
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="yautil", # Replace with your own username
    version="0.0.31",
    author="Donghwi Kim",
    author_email="dhkim09@kaist.ac.kr",
    description="Yet Another Python util.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dhkim09a/pyutil",
    packages=setuptools.find_packages(),
    # classifiers=[
    #     "Programming Language :: Python :: 3",
    #     "License :: OSI Approved :: MIT License",
    #     "Operating System :: OS Independent",
    # ],
    install_requires=deps,
    extras_require={
        'test': deps,
    },
    python_requires='>=3.7',
)
